import 'package:flutter/material.dart';

/// Gadget used to store the shared state of the application.
/// It is implemented as a [ChangeNotifier].
/// An [ObservableModel] instance is automatically stored in the [ServiceContainer] by [AppGadget]
/// during initialization.
class ObservableModel extends ChangeNotifier {
  final Map<String, Object> _beanMap = Map<String, Object>();
  final Set<String> _updatedProperties = Set<String>();

  /// Retrieves a bean from the map.
  T getBean<T>([String token]) => _beanMap[toBeanKey(T, token: token)] as T;

  /// Saves a bean reference into the map.
  /// If notify is false, it will not update its listeners.
  void putBean<T>(
    Object bean, {
    String token,
    List<String> properties,
    bool notify = true,
  }) {
    assert(bean is T, 'The provided bean is not of type $T');
    _beanMap[toBeanKey(T, token: token)] = bean;
    _updatedProperties.add(toBeanKey(T, token: token));
    if (properties != null) {
      _updatedProperties.addAll(
        properties.map(
          (property) => toBeanKey(
            T,
            token: token,
            property: property,
          ),
        ),
      );
    }
    if (notify) notifyListeners();
  }

  /// Saves multiple bean references into the map.
  void putAllBeans(Map<String, Object> beans, {bool notify = true}) {
    if (beans == null || beans.isEmpty) return;
    _beanMap.addAll(beans);
    _updatedProperties.addAll(beans.keys);
    if (notify) notifyListeners();
  }

  /// Notifies listeners the change of a given bean/property.
  /// A bean is identified using its type and the optional token.
  /// The properties parameter specifies the [List] of changed properties: in this way
  /// [ModelGadget] will update only the [SubscriberGadget]s that are observing those
  /// properties.
  ///
  /// It is used when the internal state of a bean changes but its reference in the
  /// bean map stays the same.
  void notifyFor(Type type, {String token, List<String> properties}) {
    _updatedProperties.add(toBeanKey(type, token: token));
    if (properties != null && properties.isNotEmpty) {
      properties.forEach((String currentProperty) => _updatedProperties
          .add(toBeanKey(type, token: token, property: currentProperty)));
    }
    notifyListeners();
  }

  /// Clears the [Set] of updated properties.
  void clearUpdatedProperties() => _updatedProperties.clear();

  /// Clears the beans map.
  void clearMap() => _beanMap.clear();

  /// Removes a bean from the map.
  void removeBean<T>({String token}) {
    _beanMap.remove(toBeanKey(T, token: token));
  }

  Set<String> get updatedProperties => _updatedProperties;
}

/// Converts a [Type] to it's string representation.
/// The optional parameter token can be used to represent different
/// beans with the same [Type].
/// The optional parameter property can be used to represent a bean's property.
///
/// ```dart
/// class CustomType {
///   ...
/// }
///
/// print(toBeanKey(CustomType)); // CustomType__
/// print(toBeanKey(CustomType, token: 'Token')); // CustomType_Token_
/// print(toBeanKey(CustomType, property: 'one')); // CustomType__one
/// print(toBeanKey(CustomType, token: 'Token', property: 'one')); // CustomType_Token_one
/// ```
String toBeanKey(Type type, {String token, String property}) {
  return '${type.toString()}_${token ?? ""}_${property ?? ""}';
}
